"""Algorithms which are related to US mathematician D.H. Lehmer."""

import abc
import typing

from mutwo import common_utilities

__all__ = ("Backtracking", "IndexBasedBacktracking")

Solution = tuple[typing.Any, ...]
ElementList = list[typing.Any]


class Backtracking(abc.ABC):
    """Abstract base class to implement a backtracking algorithm

    By inheriting from this class, various backtracking algorithms
    can be implemented. In order to do so the user has to override
    a set of abstract methods. The abstract methods include:

        - :abstractmethod:`Backtracking.is_valid`
        - :abstractmethod:`Backtracking.solution_count`
        - :abstractmethod:`Backtracking.append_new_element`
        - :abstractmethod:`Backtracking.update_last_element`
        - :abstractmethod:`Backtracking.can_last_element_be_updated`

    Furthermore it may be helpful to override the following method
    (even though there is a valid working implementation):

        - :method:`Backtracking.element_list_to_solution`

    Please see the methods documentation for more details.

    The implementation of this backtracking algorithm makes a distinction
    between an element list and a solution. A solution is created by an
    element list. A solution is the output a user wants to get, but
    an element list is an object which is used internally in order to
    solve the problem. When implementing a backtracking algorithm by
    using this interface the user doesn't have to make the distinction
    between both (and in this case treat both in the same way).

    The most common use case for this distinction is by having a set of
    items which can appear in the solution and a list of indices which
    item of set shall be used. In this  case the element_list is actually
    a list of indices. This use case is implemented in the
    :class:`IndexBasedBacktracking` class.

    `Bitner and Reingold [2] credit Derrick H. Lehmer with first using the
    term 'backtrack' in the 1950s.
    <https://www.chessprogramming.org/Backtracking>`_.
    """

    @property
    @abc.abstractmethod
    def solution_count(self) -> int:
        """Return expected solution size"""

    @abc.abstractmethod
    def is_valid(self, element_list: ElementList) -> bool:
        """Checks if an element list provides an acceptable solution.

        :return: `True` if the solution is acceptable and `False` if
            the solution is rejected.
        """

    @abc.abstractmethod
    def append_new_element(self, element_list: ElementList):
        """Append new element to element list.

        :param element_list: The element list to which a new
            element shall be appended.
        """

    @abc.abstractmethod
    def update_last_element(self, element_list: ElementList):
        """Increments value of the last element in an element_list.

        :param element_list: The element list which last value shall
            be updated.

        This function should raise an Exception in case the last
        element can't be updated.
        """

    @abc.abstractmethod
    def can_last_element_be_updated(self, element_list: ElementList) -> bool:
        """Checks if the last element of the list can be incremented.

        :param element_list: The element list which last value shall
            be checked.
        """

    def element_list_to_solution(self, element_list: ElementList) -> Solution:
        """Converts an element list to the final solution

        :param element_list: The element list to be converted.
        """
        return tuple(element_list)

    def solve(
        self, return_element_list: bool = False
    ) -> typing.Union[Solution, tuple[Solution, ElementList]]:
        """Apply backtracking algorithm.

        :param return_element_list: If set to `True` the function
            will not only return the solution, but also the element
            list.
        """

        element_list = []
        while True:
            if self.is_valid(element_list):
                if len(element_list) < self.solution_count:
                    self.append_new_element(element_list)
                else:
                    break
            else:
                while not self.can_last_element_be_updated(element_list):
                    element_list = element_list[:-1]
                    if len(element_list) == 0:
                        raise common_utilities.NoSolutionFoundError()

                self.update_last_element(element_list)

        solution = self.element_list_to_solution(element_list)
        if return_element_list:
            return solution, element_list
        return solution


class IndexBasedBacktracking(Backtracking):
    """Abstract base class for index based backtracking algorithms

    This class implements concrete solutions for the following methods
    which are inherited from the parent class :class:`Backtracking`:

        - :abstractmethod:`Backtracking.append_new_element`
        - :abstractmethod:`Backtracking.update_last_element`
        - :abstractmethod:`Backtracking.can_last_element_be_updated`

    The following methods still have to be implemented:

        - :abstractmethod:`Backtracking.is_valid`
        - :abstractmethod:`Backtracking.solution_count`

    (Please consult for more information the documentation
    of :class:`Backtracking`).

    Furthermore the class adds new abstract methods to be implemented
    by child classes:

        - :abstractmethod:`IndexBasedBacktracking.element_index_to_item_sequence`

    **Example:**

    >>> import itertools
    >>> from mutwo import common_generators
    >>> class QueenProblem8(common_generators.IndexBasedBacktracking):
            point_list = list(itertools.combinations_with_replacement(range(queen_count), 2))
            point_list.extend(
                [tuple(reversed(point)) for point in point_list if len(set(point)) == 2]
            )
            def element_index_to_item_sequence(self, element_index, element_list):
                return self.point_list
            @property
            def solution_count(self):
                # 8 queens problem!
                return 8
            def is_valid(self, element_list):
                solution = self.element_list_to_solution(element_list)
                for queen0, queen1 in itertools.combinations(solution, 2):
                    # x != x, y != y
                    is_valid = all(value0 != value1 for value0, value1 in zip(queen0, queen1))
                    difference_x, difference_y = (value0 - value1 for value0, value1 in zip(queen0, queen1))
                    is_valid = is_valid and (difference_x != difference_y)
                    if not is_valid: return False
                return True
    >>> queen_problem_8 = QueenProblem8()
    >>> queen_problem_8.solve()
    """

    @abc.abstractmethod
    def element_index_to_item_sequence(
        self, element_index: int, element_list: ElementList
    ) -> typing.Sequence[typing.Any]:
        """Get a sequence of items to choose from for a specific element

        :param element_index: The index of the element for which a sequence
            of solutions shall be returned.
        :param element_list: The current element list
        """

    def append_new_element(self, element_list: ElementList):
        element_list.append(0)

    def update_last_element(self, element_list: ElementList):
        element_list[-1] += 1

    def can_last_element_be_updated(self, element_list: ElementList) -> bool:
        max_index = len(
            self.element_index_to_item_sequence(len(element_list) - 1, element_list)
        )
        return element_list[-1] + 1 < max_index

    def element_list_to_solution(self, element_list: ElementList) -> Solution:
        return tuple(
            self.element_index_to_item_sequence(element_index, element_list)[element]
            for element_index, element in enumerate(element_list)
        )



"""Algorithms which are related to US-American researcher Frank Gray."""

__all__ = ("reflected_binary_code",)


def reflected_binary_code(
    length: int, modulus: int
) -> tuple[tuple[int, ...], ...]:
    """Make gray code where each tuple has `length` items with `modulus` different numbers.

    :param length: how long one code is
    :param modulus: how many different numbers are included

    **Example:**

    >>> from mutwo.generators import gray
    >>> gray.reflected_binary_code(2, 2)
    ((0, 0), (0, 1), (1, 1), (1, 0))
    >>> gray.reflected_binary_code(3, 2)
    ((0, 0, 0),
    (0, 0, 1),
    (0, 1, 1),
    (0, 1, 0),
    (1, 1, 0),
    (1, 1, 1),
    (1, 0, 1),
    (1, 0, 0))
    >>> gray.reflected_binary_code(2, 3)
    ((0, 0), (0, 1), (0, 2), (1, 2), (1, 1), (1, 0), (2, 0), (2, 1), (2, 2))

    Basic code has been copied from:
        https://yetalengthothermodulusathblog.com/tag/gray-codes/
    """

    def _recursive_gray_code_maker(
        length: int, modulus: int
    ) -> list[list[int]]:
        F = range(modulus)
        if length == 1:
            return [[i] for i in F]
        L = _recursive_gray_code_maker(length - 1, modulus)
        M: list[list[int]] = []
        for j in F:
            M = M + [ll + [j] for ll in L]
        k = len(M)
        Mr: list[list[list[int]]] = [[]] * modulus
        for i in range(modulus - 1):
            i1 = i * int(k / modulus)
            i2 = (i + 1) * int(k / modulus)
            Mr[i] = M[i1:i2]
        Mr[modulus - 1] = M[(modulus - 1) * int(k / modulus) :]  # type: ignore
        for i in range(modulus):
            if i % 2 != 0:
                Mr[i].reverse()
        M0: list[list[int]] = []
        for i in range(modulus):
            M0 = M0 + Mr[i]
        return M0

    return tuple(
        tuple(reversed(list_)) for list_ in _recursive_gray_code_maker(length, modulus)
    )



from mutwo import core_events

__all__ = (
    "InvalidMinimaCurveAndMaximaCurveCombination",
    "UnequalEnvelopeDurationError",
    "InvalidStartAtValueError",
    "NoSolutionFoundError",
)


class InvalidMinimaCurveAndMaximaCurveCombination(Exception):
    """Raise for invalid envelope combinations in :class:`mutwo.common_generators.Tendency`."""


class UnequalEnvelopeDurationError(InvalidMinimaCurveAndMaximaCurveCombination):
    def __init__(
        self, minima_curve: core_events.Envelope, maxima_curve: core_events.Envelope
    ):
        super().__init__(
            "Found unequal duration when comparing 'minima_curve' "
            f"(with duration = '{minima_curve.duration}')"
            f"  and 'maxima_curve' (with duration = "
            f"'{maxima_curve.duration}'). Make sure both curves "
            "have equal duration."
        )


class InvalidStartAtValueError(ValueError):
    """Raise for invalid error of 'start_at' in :class:`mutwo.common_generators.ActivityLevel`"""

    def __init__(self, start_at: int):
        super().__init__(
            "The value for the parameter 'start_at' has to be"
            f"assigned to either 0, 1 or 2 and not {start_at}, "
            "because there are only three different tuples defined per level."
        )


class InvalidActivityLevelError(ValueError):
    """Raise for invalid 'level' when calling :class:`mutwo.common_generators.ActivityLevel`"""

    def __init__(self, level: int, allowed_range_tuple: tuple[int, ...]):
        super().__init__(
            f"The parameter 'level' is assigned to {level} "
            f"but has to be in the range of {allowed_range_tuple}."
        )


class NoSolutionFoundError(Exception):
    """Raise in case backtracking algorithm can't find any solution"""

    def __init__(self):
        super().__init__("No solution could be found.")


class InvalidSizeError(ValueError):
    """Raise for invalid 'size' parameter value in :func:`mutwo.common_generators.paradiddle`"""

    def __init__(self, size: int):
        super().__init__(
            f"Invalid value '{size}' for argument 'size'. 'Size' has to "
            "be divisible by 2 and has to be bigger than 2."
        )



"""Algorithms which are related to German-Dutch composer G. M. Koenig."""

import numpy as np  # type: ignore
import ranges  # type: ignore

from mutwo import common_utilities
from mutwo import core_events

__all__ = ("Tendency",)


class Tendency(object):
    """Tendency offers an interface for dynamically changing minima / maxima areas.

    :param minima_curve: The curve which describes the smallest allowed value over the
        time axis.
    :type minima_curve: core_events.Envelope
    :param maxima_curve: The curve which describes the biggest allowed value over the
        time axis.
    :type maxima_curve: core_events.Envelope
    :param random_seed: The random seed which shall be set.
    :type random_seed: int

    The class is based on Gottfried Michael Koenigs algorithm of "Tendenz-Masken" in
    his program "Projekt 2" where those minima / maxima areas represent probability
    fields.

    **Example:**

    >>> import core_events
    >>> from mutwo.generators import koenig
    >>> minima_curve = core_events.Envelope.from_points((0, 0), (1, 1), (2, 0))
    >>> maxima_curve = core_events.Envelope.from_points((0, 1), (1, 2), (2, 3))
    >>> my_tendency = koenig.Tendency(minima_curve, maxima_curve)
    >>> my_tendency.value_at(0.5)
    0.6456692551041303
    >>> my_tendency.value_at(0.5)
    0.9549270045140213
    """

    def __init__(
        self,
        minima_curve: core_events.Envelope,
        maxima_curve: core_events.Envelope,
        random_seed: int = 100,
    ):
        self._assert_curves_are_valid(minima_curve, maxima_curve)
        self._minima_curve = minima_curve
        self._maxima_curve = maxima_curve
        self._random = np.random.default_rng(random_seed)

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.minima_curve}, {self.maxima_curve})"

    @staticmethod
    def _assert_curves_are_valid(
        minima_curve: core_events.Envelope, maxima_curve: core_events.Envelope
    ):
        """Helper method that asserts the curves are a valid min/max pair."""

        # make sure both curves have equal duration
        if minima_curve.duration != maxima_curve.duration:
            raise common_utilities.UnequalEnvelopeDurationError(
                minima_curve, maxima_curve
            )

        # It would be better if we could compare all local extrema.
        # But there is no public function in core_events.Envelope yet.
        # Even if there would be a function to find local extrema it
        # is uncertain if this would be faster.
        # We want to make sure that at any time point minima_curve
        # < maxima_curve.
        point_to_compare_tuple = (
            minima_curve.absolute_time_tuple
            + maxima_curve.absolute_time_tuple
            + (0, minima_curve.duration)
        )
        for time in point_to_compare_tuple:
            if not minima_curve.value_at(time) < maxima_curve.value_at(time):
                raise common_utilities.InvalidMinimaCurveAndMaximaCurveCombination(
                    f"At time '{time}' 'minima_curve' isn't smaller "
                    "than 'maxima_curve'!"
                )

    @property
    def minima_curve(self) -> core_events.Envelope:
        return self._minima_curve

    @minima_curve.setter
    def minima_curve(self, minima_curve: core_events.Envelope) -> core_events.Envelope:
        self._assert_curves_are_valid(minima_curve, self.maxima_curve)
        self._minima_curve = minima_curve

    @property
    def maxima_curve(self) -> core_events.Envelope:
        return self._maxima_curve

    @maxima_curve.setter
    def maxima_curve(self, maxima_curve: core_events.Envelope) -> core_events.Envelope:
        self._assert_curves_are_valid(self.minima_curve, maxima_curve)
        self._maxima_curve = maxima_curve

    def range_at(self, time: float) -> ranges.Range:
        """Get minima / maxima range at requested time."""
        return ranges.Range(
            self.minima_curve.value_at(time), self.maxima_curve.value_at(time)
        )

    def value_at(self, time: float) -> float:
        """Get value at requested time."""
        range_at_time = self.range_at(time)
        return self._random.uniform(range_at_time.start, range_at_time.end)



"""Algorithms which are related to US-American linguist N. Chomsky."""


import dataclasses
import functools
import operator
import typing

import treelib

from mutwo import core_utilities


__all__ = ("NonTerminal", "Terminal", "ContextFreeGrammarRule", "ContextFreeGrammar")


class NonTerminal(object):
    """Can be used as a Mixin to define context-free grammar."""


class Terminal(object):
    """Can be used as a Mixin to define context-free grammar."""


@dataclasses.dataclass
class ContextFreeGrammarRule(object):
    """Describe a context_free_grammar_rule for a :class:`ContextFreeGrammar`"""

    left_side: NonTerminal
    right_side: tuple[typing.Union[NonTerminal, Terminal], ...]


class ContextFreeGrammar(object):
    """Describe a context-free grammar and resolve non-terminals

    :param context_free_grammar_rule_sequence: A sequence of :class:`ContextFreeGrammarRule` objects. It is
        allowed to provide multiple context_free_grammar_rules with the same :attribute:`left_side`.
    :type context_free_grammar_rule_sequence: typing.Sequence[ContextFreeGrammarRule]

    This is a very reduced implementation of a context-free grammar
    which only provides the most basic functions. It is not made for
    the purpose of parsing text but rather as a technique to generate
    algorithmic data (for the sake of art creation). Therefore it is
    all about the resolution of start objects to variants of this start.
    """

    def __init__(
        self,
        context_free_grammar_rule_sequence: typing.Sequence[ContextFreeGrammarRule],
    ):
        non_terminal_list = []
        for context_free_grammar_rule in context_free_grammar_rule_sequence:
            non_terminal_list.append(context_free_grammar_rule.left_side)
            for terminal_or_non_terminal in context_free_grammar_rule.right_side:
                if isinstance(terminal_or_non_terminal, NonTerminal):
                    non_terminal_list.append(terminal_or_non_terminal)
        self._non_terminal_tuple = tuple(
            core_utilities.uniqify_sequence(non_terminal_list)
        )
        self._terminal_tuple = core_utilities.uniqify_sequence(
            tuple(
                item
                for item in functools.reduce(
                    operator.add,
                    tuple(
                        context_free_grammar_rule.right_side
                        for context_free_grammar_rule in context_free_grammar_rule_sequence
                    ),
                )
                if isinstance(item, Terminal)
            )
        )
        divided_context_free_grammar_rule_list = [[] for _ in self._non_terminal_tuple]
        for context_free_grammar_rule in context_free_grammar_rule_sequence:
            index = self._non_terminal_tuple.index(  # type: ignore
                context_free_grammar_rule.left_side
            )
            divided_context_free_grammar_rule_list[index].append(
                context_free_grammar_rule
            )
        self._divided_context_free_grammar_rule_tuple = tuple(
            tuple(context_free_grammar_rule_list)
            for context_free_grammar_rule_list in divided_context_free_grammar_rule_list
        )
        self._context_free_grammar_rule_tuple = tuple(
            context_free_grammar_rule_sequence
        )

    def _data_to_tag(
        self, data: tuple[typing.Union[NonTerminal, Terminal], ...]
    ) -> str:
        return str(data)

    def _add_node(
        self,
        tree: treelib.Tree,
        data: tuple[typing.Union[NonTerminal, Terminal], ...],
        parent: typing.Optional[treelib.Node] = None,
    ):
        tree.create_node(self._data_to_tag(data), data=data, parent=parent)

    def _resolve_content(
        self, content: tuple[typing.Union[NonTerminal, Terminal], ...]
    ) -> tuple[tuple[typing.Union[NonTerminal, Terminal], ...], ...]:
        new_data_list = []
        for nth_element, element in enumerate(content):
            if isinstance(element, NonTerminal):
                context_free_grammar_rule_tuple = (
                    self.get_context_free_grammar_rule_tuple(element)
                )
                for context_free_grammar_rule in context_free_grammar_rule_tuple:
                    data = (
                        content[:nth_element]
                        + context_free_grammar_rule.right_side
                        + content[nth_element + 1 :]
                    )
                    new_data_list.append(data)
        return tuple(new_data_list)

    @property
    def non_terminal_tuple(self) -> tuple[NonTerminal, ...]:
        return self._non_terminal_tuple  # type: ignore

    @property
    def terminal_tuple(self) -> tuple[Terminal, ...]:
        return self._terminal_tuple  # type: ignore

    @property
    def context_free_grammar_rule_tuple(self) -> tuple[ContextFreeGrammarRule, ...]:
        """Get all defined rules"""

        return self._context_free_grammar_rule_tuple

    def get_context_free_grammar_rule_tuple(
        self, non_terminal: NonTerminal
    ) -> tuple[ContextFreeGrammarRule, ...]:
        """Find all defined context_free_grammar_rules for the provided :class:`NonTerminal`.

        :param non_terminal: The left side element of the :class:`ContextFreeGrammarRule`.
        :type non_terminal: NonTerminal
        """
        index = self._non_terminal_tuple.index(non_terminal)  # type: ignore
        return self._divided_context_free_grammar_rule_tuple[index]

    def resolve_one_layer(self, tree: treelib.Tree) -> bool:
        """Resolve all leaves of the tree.

        :param tree: The tree from which all leaves should
            be resolved.
        :type tree: treelib.Tree
        :return: `True` if any leaf has been resolved and `False`
            if no resolution has happened (e.g. if there are only
            :class:`Terminal` left).
        """

        new_node = False
        for leaf in tree.leaves():
            data_tuple = self._resolve_content(leaf.data)
            for data in data_tuple:
                self._add_node(tree, data, leaf)
                new_node = True
        return new_node

    def resolve(
        self, start: NonTerminal, limit: typing.Optional[int] = None
    ) -> treelib.Tree:
        """Resolve until only :class:`Terminal` are left or the limit is reached.

        :param start: The start value.
        :type start: NonTerminal
        :param limit: The maximum node levels until the function returns a tree.
            If it is set to `None` it will only stop once all nodes are
            :class:`Terminal`.
        :type limit: typing.Optional[int]
        """

        def is_limit_reached() -> bool:
            if limit is None:
                return False
            else:
                return limit <= counter

        tree = treelib.Tree()
        self._add_node(tree, (start,))
        is_not_resolved = True
        counter = 0
        while is_not_resolved and not is_limit_reached():
            is_not_resolved = self.resolve_one_layer(tree)
            counter += 1
        return tree



from . import constants

from .brown import *
from .brun import *
from .chomksy import *
from .edwards import *
from .gray import *
from .koenig import *
from .lehmer import *
from .toussaint import *

from mutwo import core_utilities

__all__ = core_utilities.get_all(
    brown, brun, chomksy, edwards, gray, koenig, lehmer, toussaint
)

# Force flat structure
del brown, brun, core_utilities, chomksy, edwards, gray, koenig, lehmer, toussaint



"""Algorithms which are related to Norwegian mathematician V. Brun."""

import abc
import typing

import numpy as np

__all__ = ("make_bruns_euclidean_algorithm_generator",)


@typing.runtime_checkable
class _BrunEuclideanElement(typing.Protocol):
    """An ABC with abstract methods __sub__, __lt__ and __gt__."""

    __slots__ = ()

    @abc.abstractmethod
    def __sub__(self) -> typing.Any:
        pass

    @abc.abstractmethod
    def __lt__(self, other) -> bool:
        pass

    @abc.abstractmethod
    def __gt__(self, other) -> bool:
        pass


def make_bruns_euclidean_algorithm_generator(
    element_tuple: tuple[
        _BrunEuclideanElement, _BrunEuclideanElement, _BrunEuclideanElement
    ],
    matrix: np.array = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]], dtype=int),
    subtraction_index: typing.Literal[1, 2] = 1,
) -> typing.Generator:
    """Make generator which runs Bruns adaption of the Euclidean algorithm.

    :param element_tuple: The initial elements which gets re-calculated
        after each step. Type doesn't matter; objects only need to have
        the following magic methods: __sub__, __lt__ and __gt__.
    :type element_tuple: tuple[_BrunEuclideanElement, _BrunEuclideanElement, _BrunEuclideanElement]
    :param matrix: The initial matrix.
    :type matrix: np.array
    :param subtraction_index: This parameter has been added for the adaption of the
        function in
        :func:`~mutwo.generators.wilson.make_wilsons_brun_euclidean_algorithm_generator`
        and is not part of Bruns original algorithm. It describes whether in each step
        the first element gets subtracted by the second (original) or by the third
        (Wilson adaption) element.
    :type subtraction_index: typing.Literal[1, 2]

    This algorithm has been described by V. Brun in his paper "EUCLIDEAN ALGORITHMS
    AND MUSICAL THEORY" (1964).

    **Example:**

    >>> import fractions
    >>> from mutwo.generators import brun
    >>> bruns_euclidean_algorithm_generator = brun.make_bruns_euclidean_algorithm_generator(
    >>>     (
    >>>         fractions.Fraction(2, 1),
    >>>         fractions.Fraction(3, 2),
    >>>         fractions.Fraction(5, 4),
    >>>     )
    >>> )
    >>> next(bruns_euclidean_algorithm_generator)
    """

    while True:
        sorted_element_tuple = tuple(sorted(element_tuple, reverse=True))
        indices = [element_tuple.index(element) for element in sorted_element_tuple]
        matrix = matrix[indices]
        yield sorted_element_tuple, matrix.copy(), element_tuple[0]
        element_tuple = (
            sorted_element_tuple[0] - sorted_element_tuple[subtraction_index],
        ) + sorted_element_tuple[1:]
        matrix[1] += matrix[0]



import setuptools  # type: ignore

version = {}
with open("mutwo/common_version/__init__.py") as fp:
    exec(fp.read(), version)

VERSION = version["__version__"]

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

extras_require = {"testing": ["nose", "coveralls"]}

setuptools.setup(
    name="mutwo.common",
    version=VERSION,
    license="GPL",
    description="Common extension for event based framework for generative art",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Levin Eric Zimmermann",
    author_email="levin.eric.zimmermann@posteo.eu",
    url="https://github.com/mutwo-org/mutwo.common",
    project_urls={"Documentation": "https://mutwo.readthedocs.io/en/latest/"},
    packages=[
        package
        for package in setuptools.find_namespace_packages(include=["mutwo.*"])
        if package[:5] != "tests"
    ],
    setup_requires=[],
    install_requires=[
        "mutwo.core>=0.61.4, <1.00.0",
        # For the brown, brun and koenig modules
        "numpy>=1.18, <2.00",
        # For the brown module
        "scipy>=1.4.1, <2.0.0",
        # For the koenig module
        "python-ranges>=0.2.0, <1.0.0",
        # For the chomsky module
        "treelib>=1.6.1, <2.0.0",
    ],
    extras_require=extras_require,
    python_requires=">=3.9, <4",
)
